'use strict';

var schema = {};
var schemaValidation = {};
var schemaDefaults = {};

var UNDEFINED = 'undefined';
var FUNCTION = 'function';
var OBJECT = 'object';
var STRING = 'string';
var NUMBER = 'number';
var BOOLEAN = 'boolean';
var REQUIRED = 'The field "@" is required.';

/*
    @onResource {Function} :: function(name, key) return {String}
*/
function ErrorBuilder(onResource) {
	this.builder = [];
	this.onResource = onResource || null;
	this.length = 0;
	this.replacer = [];
	this.isPrepared = false;
}

function UrlBuilder() {
	this.builder = {};
}

/*
    @items {Number}
    @page {Number}
    @max {Number}
*/
function PageBuilder(items, page, max) {
	this.isNext = false;
	this.isPrev = false;
	this.items = items;
	this.count = 0;
	this.skip = 0;
	this.take = 0;
	this.page = 0;
	this.max = 0;
	this.visible = false;
	this.refresh(items, page, max);
}

/*
	Create object schema
    @name {String}
    @obj {Number}
    @defaults {Function}
    return {Object}
*/
exports.schema = function(name, obj, defaults) {

	if (typeof(obj) === UNDEFINED)
		return schema[name] || null;

	if (typeof(defaults) === FUNCTION)
		schemaDefaults[name] = defaults;

	schema[name] = obj;
	return obj;
};

exports.isJoin = function(value) {
	if (!value)
		return false;
	if (value[0] === '[')
		return true;
	return typeof(schema[value]) !== UNDEFINED;
};

/*
	Create schema validation
	@name {String},
	@arr {String Array}
	return {String Array}
*/
exports.validation = function(name, arr) {
	if (typeof(arr) === UNDEFINED)
		return schemaValidation[name] || [];

	schemaValidation[name] = arr;
	return arr;
};

/*
	Create schema object
    @name {String}
    return {Object}
*/
exports.defaults = function(name) {

	var obj = exports.schema(name);

	if (obj === null)
		return null;

	var defaults = schemaDefaults[name];
	var item = utils.extend({}, obj, true);
	var properties = Object.keys(item);
	var length = properties.length;

	for (var i = 0; i < length; i++) {

		var property = properties[i];
		var value = item[property];
		var type = typeof(value);

		if (defaults) {
			var def = defaults(property, true);
			if (typeof(def) !== UNDEFINED) {
				item[property] = def;
				continue;
			}
		}

		if (type === FUNCTION) {

			if (value === Number) {
				item[property] = 0;
				continue;
			}

			if (value === Boolean) {
				item[property] = false;
				continue;
			}

			if (value === String) {
				item[property] = '';
				continue;
			}

			if (value === Date) {
				item[property] = new Date();
				continue;
			}

			if (value === Object) {
				item[property] = {};
				continue;
			}

			if (value === Array) {
				item[property] = [];
				continue;
			}

			item[property] = value();
			continue;
		}

		if (type === NUMBER) {
			item[property] = 0;
			continue;
		}

		if (type === BOOLEAN) {
			item[property] = false;
			continue;
		}

		if (type === OBJECT) {
			item[property] = value instanceof Array ? [] : {};
			continue;
		}

		if (type !== STRING) {
			item[property] = null;
			continue;
		}

		var isArray = value[0] === '[';

		if (isArray)
			value = value.substring(1, value.length - 1);

		if (isArray) {
			item[property] = [];
			continue;
		}

		var lower = value.toLowerCase();

		if (lower.contains([STRING, 'text', 'varchar', 'nvarchar', 'binary', 'data', 'base64'])) {
			item[property] = '';
			continue;
		}

		if (lower.contains(['int', NUMBER, 'decimal', 'byte', 'float', 'double'])) {
			item[property] = 0;
			continue;
		}

		if (lower.contains('bool')) {
			item[property] = false;
			continue;
		}

		if (lower.contains(['date', 'time'])) {
			item[property] = new Date();
			continue;
		}

		if (lower.contains(['object'])) {
			item[property] = {};
			continue;
		}

		if (lower.contains(['array'])) {
			item[property] = [];
			continue;
		}

		if (lower.contains(['binary', 'data', 'base64'])) {
			item[property] = null;
			continue;
		}

		item[property] = exports.defaults(value);
	}

	return item;
};

/*
	Prepare model to schema
    @name {String}
    @model {Object}
    return {Object}
*/
exports.prepare = function(name, model) {

	var obj = exports.schema(name);

	if (obj === null)
		return null;

	var tmp;
	var item = utils.extend({}, obj, true);
	var properties = Object.keys(item);
	var defaults = schemaDefaults[name];
	var length = properties.length;

	for (var i = 0; i < length; i++) {

		var property = properties[i];
		var val = model[property];

		if (typeof(val) === UNDEFINED && defaults)
			val = defaults(property, false);

		if (typeof(val) === UNDEFINED)
			val = '';

		var value = item[property];
		var type = typeof(value);
		var typeval = typeof(val);

		if (typeval === FUNCTION)
			val = val();

		if (type === FUNCTION) {

			if (value === Number) {
				item[property] = utils.parseFloat(val);
				continue;
			}

			if (value === Boolean) {
				tmp = val.toString();
				item[property] = tmp === 'true' || tmp === '1';
				continue;
			}

			if (value === String) {
				item[property] = val.toString();
				continue;
			}

			if (value === Date) {

				tmp = null;

				switch (typeval) {
					case OBJECT:
						if (utils.isDate(val))
							tmp = val;
						else
							tmp = null;
						break;
					case NUMBER:
						tmp = new Date(val);
						break;

					case STRING:
						if (val === '')
							tmp = null;
						else
							tmp = val.parseDate();
						break;
				}

				if (tmp !== null && typeof(tmp) === OBJECT && tmp.toString() === 'Invalid Date')
					tmp = null;

				item[property] = tmp || (defaults ? isUndefined(defaults(property), null) : null);
				continue;
			}

			item[property] = isUndefined(defaults(property), null);
			continue;
		}

		if (type === NUMBER) {
			item[property] = utils.parseFloat(val);
			continue;
		}

		if (type === BOOLEAN) {
			tmp = val.toString();
			item[property] = tmp === 'true' || tmp === '1';
			continue;
		}

		if (type !== STRING) {
			item[property] = val.toString();
			continue;
		}

		var isArray = value[0] === '[';

		if (isArray)
			value = value.substring(1, value.length - 1);

		var lower = value.toLowerCase();

		if (lower.contains([STRING, 'text', 'varchar', 'nvarchar'])) {
			tmp = val.toString();

			var beg = lower.indexOf('(');
			if (beg === -1) {
				item[property] = tmp;
				continue;
			}

			var size = lower.substring(beg + 1, lower.length - 1).parseInt();
			item[property] = tmp.max(size, '...');
			continue;
		}

		if (lower.contains(['int', 'byte'])) {
			item[property] = utils.parseInt(val);
			continue;
		}

		if (lower.contains(['decimal', NUMBER, 'float', 'double'])) {
			item[property] = utils.parseFloat(val);
			continue;
		}

		if (lower.contains('bool', BOOLEAN)) {
			tmp = val.toString();
			item[property] = tmp === 'true' || tmp === '1';
			continue;
		}

		if (lower.contains(['date', 'time'])) {

			if (typeval === 'date') {
				item[property] = val;
				continue;
			}

			if (typeval === STRING) {
				item[property] = val.parseDate();
				continue;
			}

			if (typeval === NUMBER) {
				item[property] = new Date(val);
				continue;
			}

			item[property] = isUndefined(defaults(property));
			continue;
		}

		if (isArray) {

			if (!(val instanceof Array)) {
				item[property] = (defaults ? isUndefined(defaults(property, false), []) : []);
				continue;
			}

			item[property] = [];
			var sublength = val.length;
			for (var j = 0; j < sublength; j++) {
				item[property][j] = exports.prepare(value, model[property][j]);
			}

			continue;
		}

		item[property] = exports.prepare(value, model[property]);
		//item[property] = (defaults ? defaults(property, false) || null : null);
	}

	return item;
};

function isUndefined(value, def) {
	if (typeof(value) === UNDEFINED)
		return def;
	return value;
}

// ======================================================
// PROTOTYPES
// ======================================================

/*
	Add a new error
	@name {String or ErrorBuilder}
	@error {String} :: default value @ (for resources)
    return {ErrorBuilder}
*/
ErrorBuilder.prototype.add = function(name, error, path) {
	var self = this;
	self.isPrepared = false;

	if (name instanceof ErrorBuilder) {

		name.builder.forEach(function(o) {
			self.builder.push(o);
		});

		self.length = self.builder.length;
		return self;
	}

	self.builder.push({ name : name, error: error || '@', path: path });
	self.length = self.builder.length;
	return self;
};

/*
	Remove error
	@name {String}
    return {ErrorBuilder}
*/
ErrorBuilder.prototype.remove = function(name) {
	var self = this;

	self.builder = self.builder.remove(function(o) {
		return o.name === name;
	});

	self.length = self.builder.length;
	return self;
};

/*
	@name {String} :: optional, default undefined
    return {Boolean}
*/
ErrorBuilder.prototype.hasError = function(name) {
	var self = this;

	if (name) {
		return self.builder.find(function(o) {
			return o.name === name;
		}) !== null;
	}

	return self.builder.length > 0;
};

/*
	Read error message
	@name {String}
	return {String}
*/
ErrorBuilder.prototype.read = function(name) {

	var self = this;

	if (!self.isPrepared)
		self.prepare();

	var error = self.builder.find(function(o) {
		return o.name === name;
	});

	if (error !== null)
		return error.error;

	return null;
};

/*
	Clear ErrorBuilder
    return {ErrorBuilder}
*/
ErrorBuilder.prototype.clear = function() {
	var self = this;
	self.builder = [];
	self.length = 0;
	return self;
};

/*
	Add a replace rule
	@search {String}
	@newvale {String}
    return {ErrorBuilder}
*/
ErrorBuilder.prototype.replace = function(search, newvalue) {
	var self = this;
	self.isPrepared = false;
	self.replacer[search] = newvalue;
	return self;
};

/*
	Serialize ErrorBuilder to JSON format
    return {String}
*/
ErrorBuilder.prototype.json = function() {
	return JSON.stringify(this.prepare().builder);
};

/*
	Serialize ErrorBuilder to JSON format
    return {String}
*/
ErrorBuilder.prototype.JSON = function() {
	return JSON.stringify(this.prepare().builder);
};

/*
	Prepare builder with Resources
    return {ErrorBuilder}
*/
ErrorBuilder.prototype._prepare = function() {
	var self = this;

	if (self.onResource === null)
		return self;

	var builder = self.builder;
	var length = builder.length;

	for (var i = 0; i < length; i++) {

		var o = builder[i];

		if (o.error[0] !== '@')
			continue;

		if (o.error.length === 1)
			o.error = self.onResource(o.name);
		else
			o.error = self.onResource(o.error.substring(1));

		if (typeof(o.error) === UNDEFINED)
			o.error = REQUIRED.replace('@', o.name);
	}

	return self;
};

ErrorBuilder.prototype._prepareReplace = function() {

	var self = this;
	var builder = self.builder;
	var lengthBuilder = builder.length;
	var keys = Object.keys(self.replacer);
	var lengthKeys = keys.length;

	if (lengthBuilder === 0 || lengthKeys === 0)
		return self;

	for (var i = 0; i < lengthBuilder; i++) {
		var o = builder[i];
		for (var j = 0; j < lengthKeys; j++) {
			var key = keys[j];
			o.error = o.error.replace(key, self.replacer[key]);
		}
	}

	return self;
};

ErrorBuilder.prototype.prepare = function() {
	var self = this;

	if (self.isPrepared)
		return self;

	self._prepare()._prepareReplace();
	self.isPrepared = true;

	return self;
};

/*
	Refresh PageBuilder
	@items {Number}
	@page {Number}
	@max {Number}
    return {PageBuilder}
*/
PageBuilder.prototype.refresh = function(items, page, max) {
	var self = this;

	self.count = Math.floor(items / max) + (items % max > 0 ? 1 : 0);
	self.page = page - 1;

	if (self.page < 0)
		self.page = 0;

	self.items = items;
	self.skip = self.page * max;
	self.take = max;
	self.max = max;
	self.isPrev = self.page > 0;
	self.isNext = self.page < self.count - 1;
	self.visible = self.count > 1;
	self.page++;

	return self;
};

/*
	Render PageBuilder
	@fn {Function} :: function(pageIndex)
	@max {Number} :: optional, default undefined
    return {Array}
*/
PageBuilder.prototype.render = function(fn, max) {

	var self = this;
	var builder = [];

	if (typeof(max) === UNDEFINED) {
		for (var i = 1; i < self.count + 1; i++)
			builder.push(fn(i, i === self.page));
		return builder;
	}

	var half = Math.floor(max / 2);
	var pages = self.count;

	var pageFrom = self.page - half;
	var pageTo = self.page + half;
	var plus = 0;

	if (pageFrom <= 0) {
		plus = Math.abs(pageFrom);
		pageFrom = 1;
		pageTo += plus;
	}

	if (pageTo >= pages) {
		pageTo = pages;
		pageFrom = pages - max;
	}

	for (var i = pageFrom; i < pageTo + 1; i++)
		builder.push(fn(i, i === self.page));

	return builder;
};

/*
	Add parameter to UrlBuilder
	@name {String}
	@value {String}
    return {UrlBuilder}
*/
UrlBuilder.prototype.add = function(name, value) {
	var self = this;

	if (typeof(name) === 'object') {
		Object.keys(name).forEach(function(o) {
			self.builder[o] = name[o];
		});
		return;
	}

	self.builder[name] = value;
	return self;
};

/*
	Remove parameter from UrlBuilder
	@name {String}
    return {UrlBuilder}
*/
UrlBuilder.prototype.remove = function(name) {
	var self = this;
	delete self.builder[name];
	return self;
};

/*
	Read parameter
	@name {String}
    return {Object}
*/
UrlBuilder.prototype.read = function(name) {
	return this.builder[name] || null;
};

/*
	Remove all keys
    return {UrlBuilder}
*/
UrlBuilder.prototype.clear = function() {
	var self = this;
	self.builder = {};
	return self;
};

/*
	Create URL
    return {String}
*/
UrlBuilder.prototype.toString = function() {

	var self = this;
	var builder = [];

	Object.keys(self.builder).forEach(function(o) {
		builder.push(o + '=' + encodeURIComponent(self.builder[o] || ''));
	});

	return builder.join('&');
};

/*
	Has UrlBuilder values?
	@keys {String or String array}
    return {Boolean}
*/
UrlBuilder.prototype.hasValue = function(keys) {

	if (typeof(keys) === UNDEFINED)
		return false;

	var self = this;

	if (typeof(keys) === 'string')
		keys = [keys];

	for (var i = 0; i < keys.length; i++) {
		var val = self.builder[keys[i]];
		if (typeof(val) === UNDEFINED || val === null)
			return false;
	}

	return true;
};

/*
	Render parameter
	@keys {String array} :: what parameter
	@delimiter {String}
    return {String}
*/
UrlBuilder.prototype.toOne = function(keys, delimiter) {

	var self = this;
	var builder = [];

	keys.forEach(function(o) {
		builder.push(self.builder[o] || '');
	});

	return builder.join(delimiter || '&');
};

// ======================================================
// EXPORTS
// ======================================================

exports.ErrorBuilder = ErrorBuilder;
exports.PageBuilder = PageBuilder;
exports.UrlBuilder = UrlBuilder;